open Printf
open Ext
open Mn

include BnType 

(* Accessors *)
let numvars bn = Array.length bn.vars
let varname bn i = bn.vars.(i).vname
let idx v = v.idx 
let parents bn i = bn.parents.(i)
let children bn i = bn.children.(i)

let numparents bn i = List.length bn.parents.(i) 
let numchildren bn i = List.length bn.children.(i)

(* SPEED HACK:
 *   Simplify tree-structured CPDs given evidence.  This speeds up
 *   inference slightly by reducing the depth of each tree.  This
 *   modifies the BN.
 * 
 * NOTE: Resulting CPDs may not be normalized, and hence be merely
 * potential functions, not conditional probability distributions. 
 *)

let rec simplify_node v x n =
  match n with 
    Leaf a -> 
      if x.(v) < 0 || Array.length a = 1 then Leaf a
      else Leaf [|a.(x.(v))|]
  | Vertex (var,value,left,right) ->
      let l () = simplify_node v x left in
      let r () = simplify_node v x right in
      if x.(var) < 0 then
        Vertex (var, value, l(), r())
      else if x.(var) = value then
        l()
      else
        r()

let simplify bn x =
  let simplify_dist var = function
    | Table t -> Table t (* TODO: simplify tables as well...? *)
    | Tree root -> Tree (simplify_node var x root) in
  let dists' = Array.mapi simplify_dist bn.dists in
  {bn with dists=dists'}


(*
 * Get the conditional probability of a variable, given its parents.
 *)

(* Get a conditional probability from a tree-based CPD *)
let rec tree_logprob x = function
  | Leaf a -> a
  | Vertex (var, value, left, right) ->
      if x.(var) = value then 
        tree_logprob x left
      else
        tree_logprob x right 

(* Log probability of the child variable given its parents. *)
let cond_logprob bn state v =
  match bn.dists.(v) with
    Table t -> t.(state_to_pidx bn v state)
  | Tree t ->  tree_logprob state t

(* Log probability of the variable's current state (X_i = x_i), 
 * given its parents (Pi_i = pi_i) *)
let node_logscore bn state v =
  let la = cond_logprob bn state v in
  (* HACK: We have fewer probs if the network was simplified. *)
  if Array.length la = 1 then 
    la.(0) 
  else 
    la.(state.(v)) 

(* Log probability of X_i given its Markov blanket *)
let mb_logprob bn x i = 

  (* For variable i and each neighbor, compute conditional log prob
   * of each value conditioned on everything else.
   *)
  let orig_ival = x.(i) in
  let childprob ival =
    x.(i) <- ival;
    List.sumf_map (node_logscore bn x) (i :: children bn i) in
  let logprobs = Array.init (get_range bn i) childprob in
  x.(i) <- orig_ival;

  (* Normalize into the Markov blanket distribution *)
  (* TODO: Beware of underflow! *)
  let probs = Array.map exp logprobs in
  let logtotal = log (Array.sumf probs) in
  Array.map (fun p -> p -. logtotal) logprobs


(* Probabilities *)
let mb_prob bn state v = 
  Array.map exp (mb_logprob bn state v)

let cond_prob bn state v =
  Array.map exp (cond_logprob bn state v)


(* 
 * Compute the log likelihood of the network on a single example.  
 * All values must be specified -- no missing data! 
 *)
let loglikelihood bn x =
  let total = ref 0. in
  for v = 0 to numvars bn - 1 do
    assert(x.(v) >= 0);
    total := !total +. node_logscore bn x v  
  done;
  !total

let pll bn x =
  let total = ref 0. in
  for v = 0 to numvars bn - 1 do
    assert(x.(v) >= 0);
    let value = x.(v) in
    total := !total +. (mb_logprob bn x v).(value)
  done;
  !total

(* Utility function for sampling a multinomial given an array
   containing the probability of each state. *)
let sample_array a =
  assert(Array.sumf a < 1.0001 && Array.sumf a > 0.9999);
  let p = ref (Random.float 1.0) in
  let idx = ref 0 in
  while !p > a.(!idx) && !idx < Array.length a do
    p := !p -. a.(!idx);
    incr idx
  done;
  if !idx < Array.length a then 
    !idx
  else
    Array.length a - 1

(* 
 * Generate a single iid sample from the BN probability distribution. 
 *)
let sample bn =
  let x = Array.create (numvars bn) 0 in
  let sample_from_parents var =
    let v = var.idx in
    x.(v) <- sample_array (cond_prob bn x v) in
  Array.iter sample_from_parents bn.topo_vars ;
  x


(*
 * File input/output
 *)
let load_bif  = BifFile.load
let load_xmod = XmodFile.load

let output_bif  = BifFile.output
let output_xmod = XmodFile.output

(* Infer filetype from filename *)
let filename_is_xmod filename =
  Str.string_match (Str.regexp ".*\\.xmod.*") filename 0

let filename_is_bif filename =
  Str.string_match (Str.regexp ".*\\.bif.*") filename 0

(* Load a BN, inferring its filetype by its filename *)
let load_auto filename = 
  let channel = open_in filename in
  if filename_is_xmod filename then 
    load_xmod channel
  else if filename_is_bif filename then 
    load_bif channel
  else
    (* Unknown filetype.  Assume it's a .bif. *)
    load_bif channel

(* Write a BN to disk, inferring its filetype by its filename *)
let write_auto filename = 
  let channel = open_out filename in
  if filename_is_xmod filename then 
    output_xmod channel
  else if filename_is_bif filename then 
    output_bif channel
  else
    (* Unknown filetype.  Assume it's a .bif. *)
    output_bif channel


(*
 * BN --> MN methods
 *)

(* Convert a tree with distributions at the leaves to one with
   potential function values at the leaves. *)
let rec tree_to_factor cvar = function
| Leaf a ->
  let rec make_leaf i =
    let l = Mn.Factor.Leaf a.(i) in
    if i = Array.length a - 1 then l
    else Mn.Factor.Vertex(cvar, i, l, make_leaf (i+1)) in
  make_leaf 0
| Vertex(var, value, left, right) ->
  let l' = tree_to_factor cvar left in
  let r' = tree_to_factor cvar right in
  Mn.Factor.Vertex(var, value, l', r')
  
let rec table_to_factor schema cvar parents t =
  if Array.length t == 0 then 
    [||]
  else
    let len = Array.length t in
    let getval i = t.(i mod len).(i/len) in
    Array.init (Array.length t * Array.length t.(0)) getval
  
let cpd_to_factor bn cvar = function
| Tree t -> 
  Mn.Factor.Tree (tree_to_factor cvar t)
| Table t -> 
  let t' = table_to_factor (schema bn) cvar (parents bn cvar) t in
  let vars = Array.of_list (cvar :: parents bn cvar) in
  let ranges = Array.map (get_range bn) vars in
  Mn.Factor.Table (vars, ranges, t')

let to_mn bn =
  let schema = schema bn in
  let factors = Array.mapi (cpd_to_factor bn) bn.dists in
  Mn.create schema factors
