open Printf
open Ext

(* Utility sub-modules *)
module Varstate =
struct 
  include Varstate
end

module Factor =
struct 
  include Factor
end

(*
 * Type declarations
 *)

type network = {schema: int array;
                factors: Factor.factor array;
                var_to_factors: Factor.factor list array}

let numvars mn = Array.length mn.schema
let get_range mn i = mn.schema.(i) 
let num_values = get_range
let schema mn = Array.copy mn.schema
let factors mn = mn.factors

let build_var_to_factors schema factors =
  let a = Array.create (Array.length schema) [] in
  let add_factor f var = a.(var) <- f :: a.(var) in
  Array.iter (fun f -> List.iter (add_factor f) (Factor.vars f)) factors;
  a

let create schema factors =
  let var_to_factors = build_var_to_factors schema factors in
  {schema=schema; factors=factors; var_to_factors=var_to_factors}

(* Unnormalized log probability of a given state. *)
let raw_logprob mn state =
  if Array.exists (fun x -> x < 0) state then
    invalid_arg "raw_logprob"
  else 
    Array.sumf_map (fun f -> Factor.log_value f state) mn.factors

(* Probability distribution over X_i, given its Markov blanket *)
let mb_dist mn x i = 
  if Array.exists (fun x -> x < 0) x then
    invalid_arg "mb_dist";
  let orig_ival = x.(i) in
  let valsum v =
    x.(i) <- v;
    List.sumf_map (fun f -> Factor.log_value f x) mn.var_to_factors.(i) in
  let dist = LogP (Array.init mn.schema.(i) valsum) in
  x.(i) <- orig_ival;
  normalize dist


(* Pseudo-log-likelihood *)
let pll mn x = 
  if Array.length x != Array.length mn.schema then
    invalid_arg "pll"
  else begin
    let mb_logscore mn x i = 
      (logp (mb_dist mn x i)).(x.(i)) in
    let total = ref 0. in
    for i = 0 to Array.length mn.schema - 1 do
      total := !total +. mb_logscore mn x i
    done;
    !total
  end

(* Build a new MN with simpler factors, conditioned on evidence *)
let simplify mn ev =
  let schema = Array.copy mn.schema in
  let factors = Array.map (fun f -> Factor.simplify f ev) mn.factors in
  let var_to_factors = build_var_to_factors schema factors in
  {schema=schema;
   factors=factors;
   var_to_factors=var_to_factors}

let to_features mn =
  List.flatten (Array.to_list (Array.map Factor.to_features mn.factors))

let input_features_lex lexbuf =
  let fl = MnParser.featurelist MnLexer.lexer lexbuf in
  let pf2f (weight_id, weight, cond) =
    {Factor.weight_id=weight_id; Factor.weight=weight; Factor.cond=cond} in
  List.map pf2f fl

let input_features chan =
  let lexbuf = Lexing.from_channel chan in
  input_features_lex lexbuf

(* I/O *)
module MP = MnParseTypes

let load chan =
  let schema = Data.input_example chan in
  let lexbuf = Lexing.from_channel chan in
  let pmn = MnParser.mn MnLexer.lexer lexbuf in
  (* Convert the parsed factors to actual factors.
     TODO: Eventually support shared weights. *)
  let factors_l = 
    List.map (Factor.pfactor_to_factor schema) pmn.MP.factors in
  let factors = Array.of_list factors_l in 
  let var_to_factors = build_var_to_factors schema factors in
  {schema=schema;
   factors=factors;
   var_to_factors=var_to_factors}

let output out mn =
  Data.output out mn.schema;
  output_string out "MN {\n";
  Array.iter (Factor.output_factor out) mn.factors;
  output_string out "}\n"
